/** Common Kernel Checkpointer API
 * @author John Mehnert-Spahn and Eugen Feller
 *
 * Copyright (c) 2008 Heinrich-Heine-Universitaet Duesseldorf
 * All rights reserved.
 *
 * This file is part of AEM. AEM is part of XtreemOS, a Linux-based Grid
 * Operating System, see <http://www.xtreemos.eu> for more details. The
 * XtreemOS project has been developed with the financial support of the
 * European Commission's IST program under contract #FP6-033576.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *     * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *     * Neither the name of the XtreemOS project nor the names of its
 *     partners and/or contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/syscall.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <linux/connector.h>
#include <linux/netlink.h>
#include <linux/cn_proc.h>
#include <signal.h>
#include <setjmp.h>

#include <pwd.h>
#include <grp.h>
#include <security/pam_appl.h>
#include <xos_interfaces.h>
#include <xos_policy_db.h>
#include <xos_policy_mgmt.h>
#include <xos_readconf.h>
#define PAM 1

#ifdef DEBUG
#define CR_DEBUG(f, a...) printf(f, ##a)
#else
#define CR_DEBUG(f, a...) do {} while(0)
#endif

#define USE_XPAM
#define CHECKPOINT_PATH_LENGTH 		45
#define STOP_PATH_LENGTH		100

#define SIZE_PROCESS_LIST 		100
#define PROC_CHAIN_SIZE			255
#define PGR_ID_SIZE 			12
#define PROCESS_GRP_REF_ID_SEM		(PGR_ID_SIZE + 5)

#define ID_TOKEN_SIZE 			2048
#define	DN_EXT_ATTR_SIZE		1024

#define TYPE_SIZE			5
#define SPACE_SIZE			2

typedef enum {
         DISK,
         MEMORY
} media_t;

typedef enum {
	COORDINATED,
	INDEPENDENT
} cp_strategy_t;

typedef enum {
	INVALID = -1,
	PID = 0,
	PGID = 1,
	SID = 2,
	APPID = 3,
	PID_CGROUP = 4,
	PROCESS_TREE = 5
} process_group_reference_type_t;

typedef enum {
	EXECUTABLE = 0x00000000,
	SHARED_LIBRARIES = 0x00000001,
	STATICALLY_LINKED_LIBRARIES = 0x00000002,
	SPECIAL_FILES = 0x00000004,
	REGULAR_FILES = 0x0000008
} file_saving_options_t;

typedef struct aem_checkpoint_info{
	int process_group_reference_id;
	process_group_reference_type_t process_group_reference_type;
	char* pid_cgroup_name;
	media_t media;
	int* process_list;
	/* AEM does not distinguish between kernel level threads and
	 * processes, threads have same meaning as processes */
	int process_list_size;
	/* actual process number */
	int real_process_number;
	char dir_name[CHECKPOINT_PATH_LENGTH];
	cp_strategy_t strategy;
	int checkpoint_image_version;
	int incremental;
	/* NEW */
	//file_saving_options_t file_options;

}aem_checkpoint_info_t ;

typedef struct aem_restart_info{
	int process_group_reference_id;
	process_group_reference_type_t process_group_reference_type;
	char* pid_cgroup_name;
	media_t media;
	cp_strategy_t strategy;
	int checkpoint_image_version;
	int *process_list;
	int process_list_size;
	int task_identity_req_pid;
	int task_identity_req_tgid;
	char *serverPort;

}aem_restart_info_t;

struct handle_ctrl {
	char jobId[50];
	void* handle;
};

void** handle_list = NULL;

/* data structure for exchanging data */
struct credentials {
	const char *uname;
	const char *passwd;
};

struct jobExMngInfo{
	char *jobId;
	char *path;
	char *input;
	char *output;
	char *error;
	char **params;
	char **envVar;
	unsigned char * userCert;
};

/* PAM session: child-parent communication checkpoint */
int pipe_auth[2];
/* PAM session: child-parent communication restart */
int pipe_auth_rst[2];


struct return_from_authenticate{
	int ret;
	char process_group_reference_id_and_semantic[PROCESS_GRP_REF_ID_SEM];
	int uid_gid_switch;
};

struct return_from_authenticate_rst{
	int ret;
	char process_list[PROC_CHAIN_SIZE];
	char * ctid;
};

enum cp_rst_call {
	PREPARE = 0,
	STOP = 1,
	CHECKPOINT = 2,
	RESUME_CP = 3,
	REBUILD = 4,
	RESUME_RST = 5
};

/* For authentication */
uid_t g_uid = -1;
uid_t g_gid = -1;
char g_name[100];

struct jobEvent{
	char *event;
	char *jobId;
	int ppid;
	int pgid;
	int pid;
	int gid;
	int exitValue;
	int uid;
};

struct timeb {
	time_t time;
	unsigned short millitm;
	short timezone;
	short dstflag;
};

void* get_handle(const char*);
int insert_handle(void* , const char*);
void remove_handle(const char*);
